# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import numpy as np


class SumBased:
    func = np.add
    base_val = 0


class MaxBased:
    func = np.maximum
    base_val = -np.inf


class MinBased:
    func = np.minimum
    base_val = np.inf


class ProductBased:
    func = np.multiply
    base_val = 1


metrics_dict = {
    "sum": SumBased,
    "mean": SumBased,
    "max": MaxBased,
    "min": MinBased,
    "prod": ProductBased,
    "std": SumBased,
    "var": SumBased,
}
